/* 
 * Copyright (c) 2021 Satlab A/S <satlab@satlab.com>
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <endian.h>
#include <gnuradio/io_signature.h>
#include "srs3_crc32c_check_impl.h"

extern "C" {
#include "fec/crc32c.h"
#include "spaceframe.h"
}

namespace gr {
    namespace satlab {
        srs3_crc32c_check::sptr
        srs3_crc32c_check::make(bool enable)
        {
            return gnuradio::make_block_sptr<srs3_crc32c_check_impl>(enable);
        }

        void srs3_crc32c_check_impl::check(pmt::pmt_t pdu)
        {
            uint8_t frame_data[SPACEFRAME_LENGTH_MAX];
            size_t frame_length;

            /* Copy frame data to output */
            pmt::pmt_t vector = pmt::cdr(pdu);
            frame_length = pmt::blob_length(vector);
            memcpy(frame_data, pmt::blob_data(vector), frame_length);

            /* Check CRC */
            if (d_enabled) {
                uint32_t crc32, crc32_calc;
                frame_length -= sizeof(crc32);
                memcpy(&crc32, &frame_data[frame_length], sizeof(crc32));
                crc32 = be32toh(crc32);
                crc32_calc = crc32c(frame_data, frame_length);
                if (crc32 != crc32_calc)
                    return;
            }

            /* Create PMT blob */
            pmt::pmt_t vector_out = pmt::make_blob(frame_data, frame_length);
            pmt::pmt_t msg = pmt::cons(pmt::car(pdu), vector_out);

            /* Route frame to correct output */
            message_port_pub(pmt::mp("output"), msg);
        }

        void srs3_crc32c_check_impl::set_enable(bool enable)
        {
            d_enabled = enable;
        }

        srs3_crc32c_check_impl::srs3_crc32c_check_impl(bool enable)
            : gr::block("srs3_crc32c_check",
              gr::io_signature::make(0, 0, 0),
              gr::io_signature::make(0, 0, 0)),
            d_enabled(enable)
        {
            message_port_register_out(pmt::mp("output"));
            message_port_register_in(pmt::mp("check"));
            set_msg_handler(pmt::mp("check"), [this](pmt::pmt_t msg) {
                this->check(msg);
            });
        }

        srs3_crc32c_check_impl::~srs3_crc32c_check_impl()
        {
        }
    } /* namespace satlab */
} /* namespace gr */

