/* -*- c++ -*- */
/*
 * Copyright 2013 Free Software Foundation, Inc.
 *
 * This file is part of GNU Radio
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sweeping_mixer_impl.h"
#include <gnuradio/io_signature.h>

namespace gr {
namespace satlab {

sweeping_mixer::sptr sweeping_mixer::make(double sampling_rate)
{
    return gnuradio::make_block_sptr<sweeping_mixer_impl>(sampling_rate);
}

sweeping_mixer_impl::sweeping_mixer_impl(double sampling_rate)
    : sync_block("sweeping_mixer",
                 io_signature::make(1, 1, sizeof(gr_complex)),
                 io_signature::make2(1, 2, sizeof(gr_complex), sizeof(float))),
      d_sampling_rate(sampling_rate),
      d_freq_start(0),
      d_freq_stop(0),
      d_freq_step(0)
{
}

sweeping_mixer_impl::~sweeping_mixer_impl() {}

int sweeping_mixer_impl::work(int noutput_items,
                     gr_vector_const_void_star& input_items,
                     gr_vector_void_star& output_items)
{
    gr::thread::scoped_lock lock(d_mutex);
    gr_complex* input = (gr_complex*)input_items[0];
    gr_complex* output = (gr_complex*)output_items[0];

    // Optional instantaneous frequency output [rad/sample]
    float* output_freqs = nullptr;
    if (output_items.size() > 1)
        output_freqs = (float*)output_items[1];

    for (int i = 0; i < noutput_items; i++) {
        // Generate output sample = input mixed by complex sine
        output[i] = input[i] * gr_complex(d_nco.cos(), d_nco.sin());

        // Optional output
        if (output_freqs)
            output_freqs[i] = d_nco.get_freq(); // in rad/sample

        // Increment frequency by d_freq_step every sample
        d_nco.adjust_freq(d_freq_step);

        // Step phase
        d_nco.step();

        // When reaching upper or lower frequency limit, reverse sign of the frequency step taken
        if (d_nco.get_freq() >= d_freq_stop || d_nco.get_freq() <= d_freq_start)
            d_freq_step = -d_freq_step;
    }

    return noutput_items;
}

void sweeping_mixer_impl::set_sweep(double start, double stop, double rate)
{
    gr::thread::scoped_lock lock(d_mutex);

    // Hz to rad/sample
    d_freq_start = start * 2 * GR_M_PI / d_sampling_rate;
    d_freq_stop = stop * 2 * GR_M_PI / d_sampling_rate;

    // Hz/s to rad/sample^2
    d_freq_step = rate * 2.0 * GR_M_PI / d_sampling_rate / d_sampling_rate;
}

void sweeping_mixer_impl::set_frequency(double frequency)
{
    gr::thread::scoped_lock lock(d_mutex);

    // Instantaneous frequency in Hz, converted to rad/sample
    d_nco.set_freq(frequency * 2 * GR_M_PI / d_sampling_rate);
}

double sweeping_mixer_impl::get_frequency()
{
    // Instantaneous frequency in rad/sample converted to Hz
    return d_nco.get_freq() * d_sampling_rate / (2 * GR_M_PI);
}

} /* namespace satlab */
} /* namespace gr */
